#include "red_black_heuristic.h"

#include "globals.h"
#include "option_parser.h"
#include "plugin.h"
#include "state.h"
#include "causal_graph.h"
#include "../preprocess/scc.h"

#include <vector>
#include <stdint.h>
#include <stdlib.h>


using namespace std;


//#define DEBUG_RB	// compute_heuristic function
//#define DEBUG_BLACK	// computing black solution path
//#define DEBUG_RPG	// computing relaxed planning graph
//#define DEBUG_VP	// variable painting
//#define DEBUG_VP_OPS	// dump operators, too
//#define DEBUG_REAL_PLAN	// output info concerning applicability of found plan
//#define ONLY_EVAL_INIT_STATE
//#define DEBUG_RB_PLAN	// output the generated red-black plan
//#define DEBUG_CACHE
//#define DEBUG_DEAD_END



// construction and destruction
RedBlackHeuristic::RedBlackHeuristic(int max_black_size, const Options &opts)
    : FFHeuristic(false, opts), max_black_size(max_black_size) {
    preferred_ops = NONE;
}

RedBlackHeuristic::RedBlackHeuristic(const Options &opts)
    : FFHeuristic(opts) {
	if (opts.contains("max_black_size")){
	    max_black_size = opts.get<int>("max_black_size");
            if (max_black_size == -1){
                max_black_size = numeric_limits<int>::max();
            }
	} else {
	    max_black_size = 0;
	}
	if (opts.contains("preferred")){
	    preferred_ops = PreferredOpsType(opts.get_enum("preferred"));
	} else {
	    preferred_ops = NONE;
	}
}

RedBlackHeuristic::~RedBlackHeuristic(){
}

// initialization
void RedBlackHeuristic::initialize(){
    cout << "Initializing red-black heuristic..." << endl;
    
#ifndef NDEBUG
    FFHeuristic::initialize();
#endif
    
    verify_no_axioms_no_conditional_effects();
    
    if (is_unit_cost_problem()){
	cout << "using unit costs" << endl;
    }
    
    if (preferred_ops == FF){
	FFHeuristic::initialize();
	cout << "using FF's preferred operators" << endl;
    } else {
	if (preferred_ops == RB){
	    cout << "using RB preferred operators" << endl;
	}
    }
    
    assigned_to_factor = vector<int>(g_variable_domain.size(), -1);
    new_index = vector<int>(g_variable_domain.size(), -1);
    
    
        
    ///////////////////////////////////////////////////////////////// paint variables
         
    vector<vector<int> > vars = vector<vector<int> >(g_variable_domain.size());
    for (int i = 0; i < g_variable_domain.size(); i++){
        vars[i] = g_causal_graph->get_successors(i);
    }
    
    SCC scc(vars);
    vector<vector<int> > sccs = scc.get_result();    
    
#ifdef DEBUG_VP
    cout << "found " << sccs.size() << " SCCs" << endl;
#endif
    
    uint64_t current_size = 1;
    
    // contains all leaf SCCs sorted by size
    multimap<uint64_t, pair<vector<int>, set<int> > > canditate_sccs = 
                    multimap<uint64_t, pair<vector<int>, set<int> > >();

    for (int i = 0; i < sccs.size(); i++){
        set<int> scc_vars = set<int>(sccs[i].begin(), sccs[i].end());
        set<int> successor_sccs = set<int>();
        //bool containsGoal = false;
        current_size = 1;
        for (int j = 0; j < sccs[i].size(); j++){
            int var = sccs[i][j];
            current_size *= g_variable_domain[var];
            if (current_size > max_black_size){
                break;
            }
            //cout << "successors:" << endl;
            for (int succ = 0; succ < g_causal_graph->get_successors(var).size(); succ++){
                int successor = g_causal_graph->get_successors(var)[succ];
                //cout << g_fact_names[successor][0] << endl;
                if (scc_vars.find(successor) == scc_vars.end()){
                    // SCC has successor SCC
                    successor_sccs.insert(-1);
                }
            }
        }

#ifdef DEBUG_VP
        cout << "=========      SCC with " << scc_vars.size() << " variables" << endl;
        for (set<int>::iterator it = scc_vars.begin(); it != scc_vars.end(); ++it){
            cout << "   " << g_fact_names[*it][0] << endl;
        }
#endif
        if (current_size < max_black_size){
            canditate_sccs.insert(make_pair(current_size, make_pair(vector<int>(scc_vars.begin(), scc_vars.end()), successor_sccs)));
        }
#ifdef DEBUG_VP
        cout << "=========      end SCC; has size " << current_size << endl;
#endif
    }
    
#ifdef DEBUG_VP
    cout << "performing the actual factoring now, " << canditate_sccs.size() << " possible SCCs found" << endl;
#endif

    bool all_sccs_are_candidates = sccs.size() == canditate_sccs.size();
    
    // perform the actual factoring
    
    current_size = 0;
    
    vector<set<int> > current_painting;
    vector<set<int> > best_painting;
    int best_size = 0;
    bool change = true;
    while (change){
        
        if (all_sccs_are_candidates){
            if (canditate_sccs.size() <= 1){
                break;
            }
        } else {
            if (canditate_sccs.size() <= 0){
                break;
            }
        }
        
        change = false;
            
        multimap<uint64_t, pair<vector<int>, set<int> > >::iterator added_scc;

        for (multimap<uint64_t, pair<vector<int>, set<int> > >::iterator it = canditate_sccs.begin(); it != canditate_sccs.end(); ++it){
            const vector<int> &variables = it->second.first;
            set<int> &successor_sccs = it->second.second;
            int size = it->first;
            
#ifdef DEBUG_VP
            cout << "current SCC: " << variables.size() << " vars, size = " << size << endl;
            for (vector<int>::const_iterator d_it = variables.begin(); d_it != variables.end(); ++d_it){
                cout << "       " << g_fact_names[*d_it][0] << endl;
            }
#endif
            if (successor_sccs.size() == 0){
#ifdef DEBUG_VP
                cout << "is leaf SCC" << endl;
#endif
                if (current_size + size <= max_black_size){
                    current_painting.push_back(set<int>(variables.begin(), variables.end()));
                    current_size += size;
                }

                change = true;
                added_scc = it;
            } else {
#ifdef DEBUG_VP
                cout << "not a leaf SCC" << endl;
#endif
                if (successor_sccs.find(-1) != successor_sccs.end()){
#ifdef DEBUG_VP
                    cout << "has non leaf-factor successors => drop for the moment" << endl;
#endif
                    continue;
                } else {
#ifdef DEBUG_VP
                    cout << "all successors are leaf factors => combining to new leaf factor" << endl;
#endif
                    
                    int removed_size = 0;
                    
                    for (set<int>::iterator succ_it = successor_sccs.begin(); succ_it != successor_sccs.end(); ++succ_it){
                        set<int> successor = current_painting[*succ_it];
                        int tmp = 1;
                        for (set<int>::iterator succ_vars = successor.begin(); succ_vars != successor.end(); ++succ_vars){
                            if (size + current_size - removed_size - tmp > max_black_size){
                                break;
                            }
                            size *= g_variable_domain[*succ_vars];
                            tmp *= g_variable_domain[*succ_vars];
                        }
                        removed_size += tmp;
                        if (size + current_size - removed_size > max_black_size){
                            break;
                        }
                    }
                    
                    if (size + current_size - removed_size <= max_black_size){
                        
                        current_size = current_size + size - removed_size;

                        set<int> &new_factor = current_painting[*successor_sccs.begin()];
                        set<int> factors_to_remove = set<int>();
                        int k = -1;
                        for (set<int>::iterator succ_it = successor_sccs.begin(); succ_it != successor_sccs.end(); ++succ_it){
                            set<int> successor = current_painting[*succ_it];
                            if (k != -1){
                                factors_to_remove.insert(*succ_it);
                            }
                            for (set<int>::iterator scc_it = successor.begin(); scc_it != successor.end(); ++scc_it){
#ifdef DEBUG_VP
                                cout << g_fact_names[*scc_it][0] << endl;
#endif
                                new_factor.insert(*scc_it);
                            }
                            k++;
                        }
                        
                        for (int i = 0; i < variables.size(); i++){
#ifdef DEBUG_VP
                            cout << g_fact_names[variables[i]][0] << endl;
#endif
                            new_factor.insert(variables[i]);
                        }
                        
                        for (set<int>::reverse_iterator rem_it = factors_to_remove.rbegin(); rem_it != factors_to_remove.rend(); ++rem_it){
                            vector<set<int> >::iterator r_it = current_painting.begin() + *rem_it;
                            current_painting.erase(r_it);
                        }
                    }
                    
                    change = true;
                    added_scc = it; 
                }
            }
            
            if (change){
                canditate_sccs.erase(added_scc);
                
                if (current_size > best_size){
                    best_painting = current_painting;
                    best_size = current_size;
#ifdef DEBUG_VP
                    cout << "found a new best painting of size " << best_size << " with " << best_painting.size() << " WCCs" << endl;
#endif
                }
                
                // update interaction graph for remaining SCCs
                for (multimap<uint64_t, pair<vector<int>, set<int> > >::iterator it = canditate_sccs.begin(); it != canditate_sccs.end(); ++it){
                    set<int> &succ_facts = it->second.second;
                    succ_facts.clear();
                    
                    vector<int> &vars = it->second.first;
                    set<int> curr_vars = set<int>(vars.begin(), vars.end());
                    
                    for (int i = 0; i < vars.size(); i++){
                        bool found = true;
                        for (int succ = 0; succ < g_causal_graph->get_successors(vars[i]).size(); succ++){
                            found = false;
                            int successor = g_causal_graph->get_successors(vars[i])[succ];
                            if (curr_vars.find(successor) != curr_vars.end()){
                                found = true; 
                                continue;
                            }
                            for (int j = 0; j < current_painting.size(); j++){
                                if (current_painting[j].find(successor) != current_painting[j].end()){
                                    succ_facts.insert(j);
                                    found = true; 
                                    break;
                                }
                            }
                            if (!found){
                                break;
                            }
                        }
                        if (!found){
                            succ_facts.clear();
                            succ_facts.insert(-1);
                            break;
                        }
                    }
                }
                break;
            }
        }
    }
    
    // remove singleton leaves from painting
    // also filter out WCCs that don't have a goal
    bool removed = true;
    vector<bool> found_goal = vector<bool>(best_painting.size(), false);
    while (removed){
        removed = false;
        vector<set<int> >::iterator remove;
        int c = 0;
        for (vector<set<int> >::iterator it = best_painting.begin(); it != best_painting.end(); ++it){
            if (it->size() == 1){
                remove = it;
                removed = true;
                break;
            }
            if (found_goal[c]){
                c++;
                continue;
            }
            bool has_goal = false;
            for (set<int>::iterator var = it->begin(); var != it->end(); ++var){
                for (int i = 0; i < g_goal.size(); i++){
                    if (g_goal[i].first == *var){
                        has_goal = true;
                        found_goal[c] = true;
                        break;
                    }
                }
            }
            if (!has_goal){
                remove = it;
                removed = true;
            }
            c++;
        }
        if (removed){
            best_painting.erase(remove);
        }
    }
    
    cout << "black state space size " << best_size << endl;

    // perform the actual painting
    
    int number_black_vars = 0;
    for (int i = 0; i < best_painting.size(); i++){
        set<int> &variables = best_painting[i];
#ifdef DEBUG_VP
        cout << "# black vars in leaf factor " << leaf_factors.size() << ": " << variables.size() << endl;
#endif
        number_black_vars += variables.size();
        int count = 0;
        for (set<int>::iterator it = variables.begin(); it != variables.end(); ++it){
            int var = *it;
            assigned_to_factor[var] = leaf_factors.size();
#ifdef DEBUG_VP
            cout << "variable " << var << " - " << g_fact_names[var][0] << " assigned_to_factor " << leaf_factors.size() << endl;
#endif            
            new_index[var] = count++;
        }
        leaf_factors.push_back(vector<int>(variables.begin(), variables.end()));
    }
    cout << "number black variables " << number_black_vars << endl;
    cout << "number red variables " << (g_variable_domain.size() - number_black_vars) << endl;
    //}
    
#ifdef DEBUG_VP
    cout << "done painting" << endl;
#endif
    
    
    if (leaf_factors.size() == 0 && max_black_size > 0){
	cout << "no meaningful painting found, performing pure delete-relaxed search" << endl;
    }
    
    cout << "number leaf WCCs " << leaf_factors.size() << endl;
    
    ////////////////////////////////////////////////////////////////// done painting black variables
    
    curr_black_state = vector<BlackState>(leaf_factors.size(), BlackState(-1, vector<int>(), this));
    
    black_solution_cache = Cache(leaf_factors.size());
    black_solution_path = vector<vector<int> >(leaf_factors.size());
    
    black_pre_post_ops = vector<vector<int> >(leaf_factors.size());
    black_post_ops = vector<vector<int> >(leaf_factors.size());
    
    solution_found = false;    
    number_cache_hits = 0;
    
    int count = 0;
    for (int i = 0; i < g_variable_domain.size(); i++){
        if (assigned_to_factor[i] == -1){
            new_index[i] = count++;
        }
    }
    
    propositions_rb = vector<vector<RedBlackProposition> >(g_variable_domain.size(), vector<RedBlackProposition>());
    for (int var = 0; var < g_variable_domain.size(); var++) {
        propositions_rb[var] = vector<RedBlackProposition>(g_variable_domain[var]);
    }
    
    ////////////////////////////////////////////////////////////// set black operators
    for (int i = 0; i < g_operators.size(); i++){
	bool black_pre = false, black_post = false;
        int affected_factor = -1;
	for (int j = 0; j < g_operators[i].get_pre_post().size(); j++){
	    const PrePost &pre_post = g_operators[i].get_pre_post()[j];
	    if (assigned_to_factor[pre_post.var] != -1){
		black_post = true;
                affected_factor = assigned_to_factor[pre_post.var];
		if (pre_post.pre != -1){
		    black_pre = true;
		    break; // no need to loop any further, got all information
		}
	    }
	}
	if (!black_pre){
	    for (int j = 0; j < g_operators[i].get_prevail().size(); j++){
		const Prevail &prev = g_operators[i].get_prevail()[j];
		if (assigned_to_factor[prev.var] != -1){
		    black_pre = true;
		    break; // no need to loop any further, got all information
		}
	    }
	}
	if (black_pre){
	    black_pre_post_ops[affected_factor].push_back(i);
	    insert_red_pre(g_operators[i]);
	} else {
	    if (black_post){
		black_post_ops[affected_factor].push_back(i);
		insert_red_pre(g_operators[i]);
	    } else {
		red_ops.push_back(i);
	    }
	}
    }
#ifdef DEBUG_VP_OPS
    for (int factor = 0; factor < leaf_factors.size(); factor++){
        cout << "===========	number black_pre_post_ops in leaf factor " << factor << " : " << black_pre_post_ops[factor].size() << endl;
        for (int i = 0; i < black_pre_post_ops[factor].size(); i++){
            cout << "	" << g_operators[black_pre_post_ops[factor][i]].get_name() << endl;
            /*cout << "	prevails:" << endl;
            for (int j = 0; j < g_operators[black_pre_post_ops[factor][i]].get_prevail().size(); j++){
                cout << "			" << g_fact_names[g_operators[black_pre_post_ops[factor][i]].get_prevail()[j].var][g_operators[black_pre_post_ops[factor][i]].get_prevail()[j].prev] << endl;
            }
            cout << "	pre_post:" << endl;
            for (int j = 0; j < g_operators[black_pre_post_ops[factor][i]].get_pre_post().size(); j++){
                cout << "			" << (g_operators[black_pre_post_ops[factor][i]].get_pre_post()[j].pre != -1 ? g_fact_names[g_operators[black_pre_post_ops[factor][i]].get_pre_post()[j].var][g_operators[black_pre_post_ops[factor][i]].get_pre_post()[j].pre] : "none");
                cout << "	=>	" << g_fact_names[g_operators[black_pre_post_ops[factor][i]].get_pre_post()[j].var][g_operators[black_pre_post_ops[factor][i]].get_pre_post()[j].post] << endl;
            }*/
        }
        cout << "=========== number black_post_ops in leaf factor " << factor << " : " << black_post_ops[factor].size() << endl;
        for (int i = 0; i < black_post_ops[factor].size(); i++){
            cout << "	" << g_operators[black_post_ops[factor][i]].get_name() << endl;
            /*cout << "	prevails:" << endl;
            for (int j = 0; j < g_operators[black_post_ops[factor][i]].get_prevail().size(); j++){
                cout << "			" << g_fact_names[g_operators[black_post_ops[factor][i]].get_prevail()[j].var][g_operators[black_post_ops[factor][i]].get_prevail()[j].prev] << endl;
            }
            cout << "	pre_post:" << endl;
            for (int j = 0; j < g_operators[black_post_ops[factor][i]].get_pre_post().size(); j++){
                cout << "			" << (g_operators[black_post_ops[factor][i]].get_pre_post()[j].pre != -1 ? g_fact_names[g_operators[black_post_ops[factor][i]].get_pre_post()[j].var][g_operators[black_post_ops[factor][i]].get_pre_post()[j].pre] : "none");
                cout << "	=>	" << g_fact_names[g_operators[black_post_ops[factor][i]].get_pre_post()[j].var][g_operators[black_post_ops[factor][i]].get_pre_post()[j].post] << endl;
            }*/
        }
    }
    cout << "===========	number red_ops: 		" << red_ops.size() << endl;
    for (int i = 0; i < red_ops.size(); i++){
	cout << "	" << g_operators[red_ops[i]].get_name() << endl;
	/*cout << "	prevails:" << endl;
	for (int j = 0; j < g_operators[red_ops[i]].get_prevail().size(); j++){
	    cout << "			" << g_fact_names[g_operators[red_ops[i]].get_prevail()[j].var][g_operators[red_ops[i]].get_prevail()[j].prev] << endl;
	}
	cout << "	pre_post:" << endl;
	for (int j = 0; j < g_operators[red_ops[i]].get_pre_post().size(); j++){
	    cout << "			" << (g_operators[red_ops[i]].get_pre_post()[j].pre != -1 ? g_fact_names[g_operators[red_ops[i]].get_pre_post()[j].var][g_operators[red_ops[i]].get_pre_post()[j].pre] : "none");
	    cout << "	=>	" << g_fact_names[g_operators[red_ops[i]].get_pre_post()[j].var][g_operators[red_ops[i]].get_pre_post()[j].post] << endl;
	}*/
    }
    
    cout << "setting goals..." << endl;
#endif
    
    cout << "number redOps " << red_ops.size() << endl;
    int num_black_post_ops = 0, num_black_prepost_ops = 0;
    for (int i = 0; i < leaf_factors.size(); i++){
        num_black_post_ops += black_post_ops[i].size();
        num_black_prepost_ops += black_pre_post_ops[i].size();
    }
    cout << "number blackPostOps " << num_black_post_ops << endl;
    cout << "number blackPrePostOps " << num_black_prepost_ops << endl;
    
    // set goals
    red_goals = vector<pair<int, int> >();
    black_goals = vector<vector<pair<int, int> > >(leaf_factors.size());
    for (int i = 0; i < g_goal.size(); i++){
        int affected_factor = assigned_to_factor[g_goal[i].first];
        if (affected_factor != -1){
            black_goals[affected_factor].push_back(g_goal[i]);
        } else {
            red_goals.push_back(g_goal[i]);
        }
    }
    
#ifdef DEBUG_VP    
    cout << "===========        red goals:" << endl;
    for (vector<pair<int, int> >::iterator it = red_goals.begin(); it != red_goals.end(); ++it){
        cout << g_fact_names[it->first][it->second] << endl;
    }
    cout << "===========        black goals:" << endl;
    for (int i = 0; i < black_goals.size(); i++){
        for (int j = 0; j < black_goals[i].size(); j++){
            cout << g_fact_names[black_goals[i][j].first][black_goals[i][j].second] << endl;
        }
    }
#endif

    // Build goal propositions.
    for (int i = 0; i < red_goals.size(); i++) {
        int var = red_goals[i].first;
        int val = red_goals[i].second;
        goal_propositions.push_back(&propositions_rb[var][val]);
    }

    //////////////////////////////////          build unary operators
    for (int i = 0; i < red_ops.size(); i++){
        build_relaxed_operator(red_ops[i], -1);
    }
    for (int i = 0; i < black_post_ops.size(); i++){
        for (int j = 0; j < black_post_ops[i].size(); j++){
            int op_no = black_post_ops[i][j];
            if (contains_red_post(g_operators[op_no])){
                build_relaxed_operator(op_no, i);
            }
        }
    }
    
    all_relaxed_ops = relaxed_operators;
    
    relaxed_operators = vector<RedBlackOperator>();
    
    // Build unary operators for operators
    for (int i = 0; i < red_ops.size(); i++){
        build_relaxed_operator(red_ops[i], -1);
    }
    
    red_relaxed_ops = relaxed_operators;
    all_ops_referenced = false;

    reference_ops();
    
    cout << "initialization of red-black heuristic done [" << g_timer << "]" << endl;
}

void RedBlackHeuristic::build_relaxed_operator(int op_index, int affected_factor) {
    const Operator &op = g_operators[op_index];
    const vector<Prevail> &prevail = op.get_prevail();
    const vector<PrePost> &pre_post = op.get_pre_post();
    vector<RedBlackProposition *> precondition;
    vector<RedBlackProposition *> effects;
    for (int i = 0; i < prevail.size(); i++){
        precondition.push_back(&propositions_rb[prevail[i].var][prevail[i].prev]);
    }
    for (int i = 0; i < pre_post.size(); i++) {
        if (pre_post[i].pre != -1){
            precondition.push_back(&propositions_rb[pre_post[i].var][pre_post[i].pre]);
        }
        effects.push_back(&propositions_rb[pre_post[i].var][pre_post[i].post]);
    }
    RedBlackOperator relaxed_op(precondition, effects, op_index, get_adjusted_cost(op), affected_factor);
    if (relaxed_op.precondition.empty()){
        relaxed_op.precondition.push_back(&artificial_precondition);
    }
    relaxed_operators.push_back(relaxed_op);
}

void RedBlackHeuristic::insert_red_pre(const Operator& black_op) {
    
    for (int j = 0; j < black_op.get_pre_post().size(); j++){
	const PrePost &pre_post = black_op.get_pre_post()[j];
	if (assigned_to_factor[pre_post.var] == -1){
	    if (pre_post.pre != -1){
		goal_propositions.push_back(&propositions_rb[pre_post.var][pre_post.pre]);
	    }
	}
    }
    for (int j = 0; j < black_op.get_prevail().size(); j++){
	const Prevail &prev = black_op.get_prevail()[j];
	if (assigned_to_factor[prev.var] == -1){
            goal_propositions.push_back(&propositions_rb[prev.var][prev.prev]);
	}
	
    }
}

bool RedBlackHeuristic::red_goals_achieved() {
    for (int i = 0; i < red_goals.size(); i++){
        int var = red_goals[i].first;
        int val = red_goals[i].second;
        if (propositions_rb[var][val].status != REACHED){
            return false;
        }
    }
    return true;
}

bool RedBlackHeuristic::contains_red_post(const Operator &op) {
    for (int j = 0; j < op.get_pre_post().size(); j++){
	if (assigned_to_factor[op.get_pre_post()[j].var] == -1){
	    return true;
	}
    }
    return false;
}

bool RedBlackHeuristic::is_applicable_ignore_black(const Operator &op) {
    for (int j = 0; j < op.get_prevail().size(); j++){
        int var = op.get_prevail()[j].var;
	if (assigned_to_factor[var] == -1){
            int val = op.get_prevail()[j].prev;
            //if (!rpg_facts[new_index[op.get_prevail()[j].var]][op.get_prevail()[j].prev].is_true){
            if (propositions_rb[var][val].status != REACHED){
                return false;
            }
	}
    }
    for (int j = 0; j < op.get_pre_post().size(); j++){
        int var = op.get_pre_post()[j].var;
	if (assigned_to_factor[op.get_pre_post()[j].var] == -1 && op.get_pre_post()[j].pre != -1){
            //if (!rpg_facts[new_index[op.get_pre_post()[j].var]][op.get_pre_post()[j].pre].is_true){
            int val = op.get_pre_post()[j].pre;
            if (propositions_rb[var][val].status != REACHED){
                return false;
            }
	}
    }
    return true;
}

bool RedBlackHeuristic::black_goal_reached(const BlackState& curr_state, int leaf_factor) {
    for (vector<pair<int, int> >::iterator it = black_goals[leaf_factor].begin(); it != black_goals[leaf_factor].end(); ++it){
	if (curr_state[it->first] != it->second){
	    return false;
	}
    }
    return true;
}

bool RedBlackHeuristic::achieve_black_goals(const State &state, bool all_black_ops) {
    
#ifdef DEBUG_BLACK
    cout << "===========	starting search in black DTG" << endl;
    for (int i = 0; i < curr_black_state.size(); i++){
        curr_black_state[i].dump();
    }
#endif

    bool all_goals_reached = true;

    for (int factor = 0; factor < leaf_factors.size(); factor++){
    
        deque<OpenListEntry> open_list = deque<OpenListEntry>();
        open_list.push_back(OpenListEntry(curr_black_state[factor], deque<int>()));
        
        __gnu_cxx::hash_set<int> seen_states = __gnu_cxx::hash_set<int>();
        seen_states.insert(get_hash(curr_black_state[factor], factor));
        
        BlackState current_state = open_list.front().state;
        deque<int> current_path = open_list.front().path;    
        
        if (all_black_ops && useful_black_post_ops[factor].size() > 0){	// currBlackState is not valid => apply all permutations of usefulBlackPostOps
            seen_states.erase(get_hash(curr_black_state[factor], factor));
            deque<pair<BlackState, deque<int> > > useful_ops_open_list = deque<pair<BlackState, deque<int> > >();
            useful_ops_open_list.push_back(make_pair(curr_black_state[factor], deque<int>()));
            open_list.pop_back();	// remove currentBlackState because it is not valid any more
            while (!useful_ops_open_list.empty()){
                current_state = useful_ops_open_list.front().first;
                current_path = useful_ops_open_list.front().second;
                useful_ops_open_list.pop_front();
            
                for (int i = 0; i < useful_black_post_ops[factor].size(); i++){
		    const Operator &op = g_operators[useful_black_post_ops[factor][i]];
                    if (is_applicable_ignore_black(op)){
                        BlackState succ = current_state.getSuccessor(op);
                        if (seen_states.insert(get_hash(succ, factor)).second){
                            deque<int> new_path = deque<int>(current_path);
                            new_path.push_back(useful_black_post_ops[factor][i]);
                            open_list.push_back(OpenListEntry(succ, new_path));
                        }
                    }
                }
            }
        }
        
        bool goal_reached = false;
        
        while (!open_list.empty()){
            current_state = open_list.front().state;
            current_path = open_list.front().path;
            open_list.pop_front();
            
#ifdef DEBUG_BLACK
                cout << "===========	current search state: " << endl;
                current_state.dump();
#endif    
            
            if (black_goal_reached(current_state, factor)){
                    
#ifdef DEBUG_BLACK
                cout << "===========	black goal reached" << endl;
#endif
                goal_reached = true;
                break;
            }
            
#ifdef DEBUG_CACHE
            cout << "--- checking cache" << endl;
#endif
            vector<int> cached_path = black_solution_cache.get(get_hash(current_state, factor), factor);
            if (cached_path.size() != 0){
                bool contains_black_post_op = false;
                bool applicable = true;
                for (int i = 0; i < cached_path.size(); i++){
                    bool hasBlackPre = false;
                    Operator &op = g_operators[cached_path[i]];
                    for (int prev = 0; prev < op.get_prevail().size(); prev++){
                        int var = op.get_prevail()[prev].var;
                        if (assigned_to_factor[var] != -1){
                            hasBlackPre = true;
                        } else {
                            //if (!rpg_facts[new_index[op.get_prevail()[prev].var]][op.get_prevail()[prev].prev].is_true){
                            int val = op.get_prevail()[prev].prev;
                            if (propositions_rb[var][val].status != REACHED){
                                applicable = false;
                                break;
                            }
                        }
                    }
                    if (applicable){
                        for (int pre = 0; pre < op.get_pre_post().size(); pre++){
                            int var = op.get_pre_post()[pre].var;
                            if (assigned_to_factor[var] != -1){
                                hasBlackPre = true;
                            } else {
                                if (op.get_pre_post()[pre].pre != -1){
                                    int val = op.get_pre_post()[pre].pre;
                                    //if (!rpg_facts[new_index[op.get_pre_post()[pre].var]][op.get_pre_post()[pre].pre].is_true){
                                    if (propositions_rb[var][val].status != REACHED){
                                        applicable = false;
                                        break;
                                    }
                                }
                            }
                        }
                        if (!hasBlackPre){
                            contains_black_post_op = true;
                        }
                    }
                }
                
                if (applicable && (!contains_black_post_op || (contains_black_post_op == all_black_ops))){
                
#ifdef DEBUG_CACHE
                    cout << "###### cache hit for following state: size of cached path = " << cached_path.size() << endl;
                    current_state.dump();
#endif
                    current_path.insert(current_path.end(), cached_path.begin(), cached_path.end());
                    
#ifdef DEBUG_CACHE
                    cout << "    ## complete path:" << endl;
                    for (deque<int>::iterator iter = current_path.begin(); iter != current_path.end(); ++iter){
                        cout << "	" << g_operators[*iter].get_name() << endl;
                    }
#endif

                    goal_reached = true;
                    number_cache_hits++;
                    break;
                }
            }
#ifdef DEBUG_CACHE
            cout << "--- current state not in cache" << endl;
#endif	
            
            if (all_black_ops){
                for (int i = 0; i < black_post_ops[factor].size(); i++){
#ifdef DEBUG_BLACK
                    cout << "=== try to apply action " << g_operators[black_post_ops[factor][i]].get_name() << endl;
#endif
                    if (is_applicable_ignore_black(g_operators[black_post_ops[factor][i]])){
                        BlackState succ = current_state.getSuccessor(g_operators[black_post_ops[factor][i]]);
#ifdef DEBUG_BLACK
                        cout << "	is applicable" << endl;
                        cout << "generated successor: " << endl;
                        succ.dump();
#endif
                        if (seen_states.insert(get_hash(succ, factor)).second){
#ifdef DEBUG_BLACK
                            cout << "added to openList" << endl;
#endif
                            deque<int> new_path = deque<int>(current_path);
                            new_path.push_back(black_post_ops[factor][i]);
                            open_list.push_back(OpenListEntry(succ, new_path));
                        }
                    }
                }
            }

            for (int i = 0; i < black_pre_post_ops[factor].size(); i++){
#ifdef DEBUG_BLACK
                cout << "=== try to apply action " << g_operators[black_pre_post_ops[factor][i]].get_name() << endl;
#endif
                if (current_state.isApplicable(g_operators[black_pre_post_ops[factor][i]], propositions_rb)){
                    BlackState succ = current_state.getSuccessor(g_operators[black_pre_post_ops[factor][i]]);
#ifdef DEBUG_BLACK
                    cout << "	is applicable" << endl;
                    cout << "generated successor: " << endl;
                    succ.dump();
#endif
                    if (seen_states.insert(get_hash(succ, factor)).second){
#ifdef DEBUG_BLACK
                        cout << "added to openList" << endl;
#endif
                        deque<int> new_path = deque<int>(current_path);
                        new_path.push_back(black_pre_post_ops[factor][i]);
                        open_list.push_back(OpenListEntry(succ, new_path));
                    }
                }
            }
        }
        

        if (goal_reached){
            
#if defined DEBUG_BLACK || defined DEBUG_RB_PLAN
            cout << "======	found solution path: length = " << current_path.size() << endl;
#endif	
            black_solution_path[factor] = vector<int>(current_path.begin(), current_path.end());
            deque<int>::iterator it = current_path.begin();
            current_state = curr_black_state[factor];
            
            for (int i = 0; i < current_path.size(); i++){
                        
#ifdef DEBUG_CACHE
                cout << "=== adding path to cache for state:" << endl;
                current_state.dump();
                cout << "  = path:" << endl;
                for (deque<int>::iterator iter = it; iter != current_path.end(); ++iter){
                    cout << "	" << g_operators[*iter].get_name() << endl;
                }
                cout << "  = end path" << endl;
#endif
                Operator &op = g_operators[current_path[i]];
                
                black_solution_cache.insertPath(get_hash(current_state, factor), factor, vector<int>(it, current_path.end()));
                current_state.apply(op);
                ++it;

                        
#if defined DEBUG_RB_PLAN || defined DEBUG_BLACK
                cout << "	" << g_operators[current_path[i]].get_name() << endl;
#endif	    
                
                for (int prev = 0; prev < op.get_prevail().size(); prev++){
                    int var = op.get_prevail()[prev].var;
                    if (assigned_to_factor[var] == -1){
                        int val = op.get_prevail()[prev].prev;
                        if (state[var] != val){
                            if (!propositions_rb[var][val].marked){
                                propositions_rb[var][val].marked = true;
#ifdef DEBUG_BLACK
                                cout << "	needs red prevail: " << g_fact_names[op.get_prevail()[prev].var][op.get_prevail()[prev].prev] << endl;
#endif
                            }
                        }	// only new redfacts need to be achieved by RPG
                    }
                }
                for (int pre = 0; pre < op.get_pre_post().size(); pre++){
                    int var = op.get_pre_post()[pre].var;
                    if (assigned_to_factor[var] == -1){
                        if (op.get_pre_post()[pre].pre != -1){
                            int val = op.get_pre_post()[pre].pre;
                            if (state[var] != val){
                                if (!propositions_rb[var][val].marked){
                                    propositions_rb[var][val].marked = true;
#ifdef DEBUG_BLACK
                                    cout << "	needs red precondition: " << g_fact_names[op.get_pre_post()[pre].var][op.get_pre_post()[pre].pre] << endl;
#endif
                                }
                            }
                        }	// only new redfacts need to be achieved by RPG
                    }
                }
            }
#ifdef DEBUG_RB_PLAN
                cout << "======	end black solution path" << endl;
#endif
        } else {
            all_goals_reached = false;
            break;
        }        
    }
    
    if (all_goals_reached){
        return true;
    }
    
#ifdef DEBUG_BLACK
    cout << "===========	black DTG search ended without finding path to goal" << endl;
#endif
    
    // TODO store dead end states in a set to not handle them multiple times
    // it's not that easy, as new red facts make different black values reachable..
    
    return false;
}

int RedBlackHeuristic::get_rpg_cost(const State &state){
    
#ifdef DEBUG_RPG
    cout << "_______________________________________________" << endl;
    cout << "start computing rpgcost" << endl;
#endif
    
    set<const Operator*> needed_ops;
    int cost = 0;
    
    vector<vector<bool> > handled_props(propositions_rb.size());
    for (int i = 0; i < handled_props.size(); i++){
        handled_props[i].resize(propositions_rb[i].size(), false);
    }

    bool change = true;
    while (change){
	change = false;
	for (int var = 0; var < propositions_rb.size(); var++){
            if (assigned_to_factor[var] != -1){
                continue;
            }
	    for (int val = 0; val < propositions_rb[var].size(); val++){
                
		if (propositions_rb[var][val].marked && !handled_props[var][val]){
		    
		    handled_props[var][val] = true;
		    
#ifdef DEBUG_RPG
		    cout << "handled " << g_fact_names[var][val] << endl;
#endif
                    int op_index = propositions_rb[var][val].achieved_by;
		    
		    if (op_index == -1){	// proposition was true in current state
			continue;
		    }
		    
                    const Operator *op = &g_operators[op_index];
                    
		    if (!needed_ops.insert(op).second){	// op already chosen?
			continue;
		    }
   
		    cost += get_adjusted_cost(*op);
		    
#ifdef DEBUG_RPG
		    cout << "current fact: " << g_fact_names[var][val] << endl;
		    cout << "operator: " << op->get_name() << endl;
		    cout << "	number prevail: " << op->get_prevail().size() << endl;
#endif
		    for (int j = 0; j < op->get_prevail().size(); j++){
                        int variable = op->get_prevail()[j].var;
                        int value = op->get_prevail()[j].prev;
#ifdef DEBUG_RPG
			cout << g_fact_names[variable][value] << endl;
#endif
			if (state[variable] == value){
#ifdef DEBUG_RPG
			    cout << "	true in current state" << endl;
#endif
			    continue;	// no need to achieve facts that are true in current state already
			}

			// mark prevail of op \cup rpgFacts[k]
			if (!handled_props[variable][value]){
#ifdef DEBUG_RPG
			    cout << "	not yet handled" << endl;
#endif
                            propositions_rb[variable][value].marked = true;
			    change = true;
			}
#ifdef DEBUG_RPG
			cout << "	marked" << endl;
#endif
		    }
#ifdef DEBUG_RPG
		    cout << "	number prepost: " << op->get_pre_post().size() << endl;
#endif
		    for (int j = 0; j < op->get_pre_post().size(); j++){
                        int value = op->get_pre_post()[j].pre;
			if (value == -1){
#ifdef DEBUG_RPG
			    cout << "	effect only" << endl;
#endif
			    continue;
			}
			int variable = op->get_pre_post()[j].var;
#ifdef DEBUG_RPG
			cout << g_fact_names[variable][value] << endl;
#endif
			if (state[variable] == value){
#ifdef DEBUG_RPG
			    cout << "	true in current state" << endl;
#endif
                            continue;   // no need to achieve facts that are true in current state already
                        }
			// mark precondition of op \cup rpgFacts[k]
			if (!handled_props[variable][value]){
#ifdef DEBUG_RPG
			    cout << "	not yet handled" << endl;
#endif
                            propositions_rb[variable][value].marked = true;
			    change = true;
			}
#ifdef DEBUG_RPG
			cout << "	marked" << endl;
#endif
		    }
		}
	    }
	}
    }
    
    red_solution_path.clear();
    red_solution_path.reserve(needed_ops.size());
    vector<vector<bool> > red_state(g_variable_domain.size());
    for (size_t var = 0; var < g_variable_domain.size(); ++var){
        if (assigned_to_factor[var] == -1){
            red_state[var].resize(g_variable_domain[var], false);
            red_state[var][state[var]] = true;
        }
    }
    while (!needed_ops.empty()){        // sequence red ops so they are all (red-)applicable
        const Operator *op;
        bool applicable = true;
        for (const auto &o : needed_ops){
            applicable = true;
            const vector<Prevail> &prevail = o->get_prevail();
            for (int j = 0; j < prevail.size(); j++){
                int var = prevail[j].var;
                if (assigned_to_factor[var] == -1){
                    int val = prevail[j].prev;
                    if (!red_state[var][val]){
                        applicable = false;
                        break;
                    }
                }
            }
            const vector<PrePost> &prepost = o->get_pre_post();
            for (int j = 0; j < prepost.size(); j++){
                int val = prepost[j].pre;
                if (val == -1){
                    continue;
                }
                int var = prepost[j].var;
                if (assigned_to_factor[var] == -1){
                    if (!red_state[var][val]){
                        applicable = false;
                        break;
                    }
                }
            }
            
            if (applicable){
                for (int j = 0; j < prepost.size(); j++){
                    int var = prepost[j].var;
                    if (assigned_to_factor[var] == -1){
                        int eff = prepost[j].post;
                        red_state[var][eff] = true;
                    }
                }
                op = o;
                break;
            }
        }
        
        assert(applicable);
        red_solution_path.push_back(op);
        needed_ops.erase(op);
    }
#if defined DEBUG_RPG || defined DEBUG_RB_PLAN
    cout << "======		actions found in rpg" << endl;
    for (int i = 0; i < red_solution_path.size(); i++){
	cout << "	" << red_solution_path[i]->get_name() << endl;
    }
    cout << "end calculating rpg cost = " << cost << endl;
    cout << "_______________________________________________" << endl;
#endif
    
    return cost;
}

int RedBlackHeuristic::get_hash(const BlackState &state, int leaf_factor){
    int Ni = 1;
    int hash = 0;
    for (int i = 0; i < leaf_factors[leaf_factor].size(); i++){
	hash += Ni * state[leaf_factors[leaf_factor][i]];
	Ni *= g_variable_domain[leaf_factors[leaf_factor][i]];
    }
    return hash;
}

int RedBlackHeuristic::get_black_solution_path_cost(){
    int cost = 0;
    for (int i = 0; i < black_solution_path.size(); i++){
        for (int j = 0; j < black_solution_path[i].size(); j++){
            cost += get_adjusted_cost(g_operators[black_solution_path[i][j]]);
        }
    }
    return cost;
}

bool RedBlackHeuristic::is_real_plan(const State &state) const {
    
#ifdef DEBUG_REAL_PLAN
    cout << "checking if found RB plan is also real plan" << endl;
#endif
    vector<int> curr_state = vector<int>(g_variable_domain.size());
    int cost = 0;
    for (int i = 0; i < g_variable_domain.size(); i++){	// init current state
	curr_state[i] = state[i];
    }
    int plan_length = red_solution_path.size();
    vector<int> overall_black_solution_path;
    for (int i = 0; i < leaf_factors.size(); i++){
        plan_length += black_solution_path[i].size();
        overall_black_solution_path.insert(overall_black_solution_path.end(), 
                                           black_solution_path[i].begin(), black_solution_path[i].end());
    }
    
    for (int i = 0; i < plan_length; i++){	// first apply red actions
	const Operator *op;
        if (i < red_solution_path.size()){
            op = red_solution_path[i];
        } else {
            op = &g_operators[overall_black_solution_path[i - red_solution_path.size()]];
        }
	cost += get_adjusted_cost(*op);
#ifdef DEBUG_REAL_PLAN
	cout << "applying operator " << op->get_name() << endl;
#endif
	for (int prev = 0; prev < op->get_prevail().size(); prev++){	// check prevail
	    if (curr_state[op->get_prevail()[prev].var] != op->get_prevail()[prev].prev){
#ifdef DEBUG_REAL_PLAN
		cout << "op not applicable, missing prevail:" << endl;
		cout << "	" << g_fact_names[op->get_prevail()[prev].var][op->get_prevail()[prev].prev] << endl;
#endif
		return false;
	    }
	}
	for (int pre = 0; pre < op->get_pre_post().size(); pre++){	// check pre
	    if (op->get_pre_post()[pre].pre != -1 && curr_state[op->get_pre_post()[pre].var] != op->get_pre_post()[pre].pre){
#ifdef DEBUG_REAL_PLAN
		cout << "op not applicable, missing pre:" << endl;
		cout << "	" << g_fact_names[op->get_pre_post()[pre].var][op->get_pre_post()[pre].pre] << endl;
#endif
		return false;
	    }
	}
	for (int post = 0; post < op->get_pre_post().size(); post++){	// apply post
	    curr_state[op->get_pre_post()[post].var] = op->get_pre_post()[post].post;
	}
    }
    for (int i = 0; i < g_goal.size(); i++){	// check goal
	if (curr_state[g_goal[i].first] != g_goal[i].second){
#ifdef DEBUG_REAL_PLAN
		cout << "goal not reached:" << endl;
		cout << "	" << g_fact_names[g_goal[i].first][g_goal[i].second] << endl;
		cout << "but is	" << g_fact_names[g_goal[i].first][curr_state[g_goal[i].first]] << endl;
#endif
	    return false;
	}
    }
    cout << "found RB plan is also real plan of cost " << cost << endl;
    return true;
}

void RedBlackHeuristic::check_plan_applicability(const State &state){
    assert(is_redblack_plan(state));
    if (is_real_plan(state)){   // TODO include the check in plan extraction phase, so this has not to
	solution_found = true;  // be computed separately every time!
	suffix_plan = vector<const Operator*>(red_solution_path.size());
	for (int i = 0; i < red_solution_path.size(); i++){
	    suffix_plan[i] = red_solution_path[i];
	}
	for (int j = 0; j < black_solution_path.size(); j++){
            for (int i = 0; i < black_solution_path[j].size(); i++){
                suffix_plan.push_back(&g_operators[black_solution_path[j][i]]);
            }
        }
	cout << "number cache hits " << number_cache_hits << endl;
    } else {
	// set preferred operator(s)
	if (preferred_ops == RB){
	    for (int i = 0; i < red_solution_path.size(); i++){
		if (red_solution_path[i]->is_applicable(state)){
		    set_preferred(red_solution_path[i]); // set all applicable red actions preferred
		    // it currently makes no sense to mark red ops
		}
	    }
	    for (int i = 0; i < black_solution_path.size(); i++){
                for (int j = 0; j < black_solution_path[i].size(); j++){
                    if (g_operators[black_solution_path[i][j]].is_applicable(state)){
                        set_preferred(&g_operators[black_solution_path[i][j]]);
                    }
                }
	    }
	} else if (preferred_ops == FF){
            // use FF's marking
            FFHeuristic::compute_heuristic(state);
	}
    }
}

bool RedBlackHeuristic::found_solution() {
    return solution_found;
}

void RedBlackHeuristic::relaxed_exploration(const State &state) {
    priority_queue.clear();
    
    for (int var = 0; var < propositions_rb.size(); var++) {
        for (int value = 0; value < propositions_rb[var].size(); value++) {
            RedBlackProposition &prop = propositions_rb[var][value];
            prop.status = UNREACHED;
            prop.marked = false;
            prop.is_goal = false;
            prop.achieved_by = 0;
            prop.h_max_cost = numeric_limits<int>::max();
        }
    }
    int number_goals = goal_propositions.size();
    for (int i = 0; i < goal_propositions.size(); i++){
        if (i >= goal_propositions.size() - red_goals.size()){
            goal_propositions[i]->marked = true; // first entries are black op's red preconditions
        }
        goal_propositions[i]->is_goal = true;
    }
    
    artificial_precondition.status = UNREACHED;

    for (int i = 0; i < relaxed_operators.size(); i++) {
        RedBlackOperator &op = relaxed_operators[i];
        op.unsatisfied_preconditions = op.precondition.size();
        op.h_max_supporter = 0;
        op.h_max_supporter_cost = numeric_limits<int>::max();
    }

    for (int var = 0; var < propositions_rb.size(); var++) {
        RedBlackProposition *init_prop = &propositions_rb[var][state[var]];
        enqueue_if_necessary(init_prop, 0, 0);
    }
    enqueue_if_necessary(&artificial_precondition, 0, 0);    
    
    while (!priority_queue.empty()) {
        pair<int, RedBlackProposition *> top_pair = priority_queue.pop();
        int popped_cost = top_pair.first;
        RedBlackProposition *prop = top_pair.second;
        int prop_cost = prop->h_max_cost;
        assert(prop_cost <= popped_cost);
        if (prop_cost < popped_cost)
            continue;
        const vector<RedBlackOperator *> &triggered_operators =
            prop->precondition_of;
        for (int i = 0; i < triggered_operators.size(); i++) {
            RedBlackOperator *relaxed_op = triggered_operators[i];
            relaxed_op->unsatisfied_preconditions--;
            assert(relaxed_op->unsatisfied_preconditions >= 0);
            if (relaxed_op->unsatisfied_preconditions == 0) {
                relaxed_op->h_max_supporter = prop;
                relaxed_op->h_max_supporter_cost = prop_cost;
                int target_cost = prop_cost + relaxed_op->cost;
                for (int j = 0; j < relaxed_op->effects.size(); j++) {
                    RedBlackProposition *effect = relaxed_op->effects[j];
                    enqueue_if_necessary(effect, target_cost, relaxed_op);
                    if (effect->is_goal){
                        effect->is_goal = false;
                        if (--number_goals == 0){
                            return;
                        }
                    }
                }
            }
        }
    }
}

void RedBlackHeuristic::reference_ops() {
    for (int i = 0; i < relaxed_operators.size(); i++) {
        RedBlackOperator *op = &relaxed_operators[i];
        for (int j = 0; j < op->precondition.size(); j++)
            op->precondition[j]->precondition_of.push_back(op);
        for (int j = 0; j < op->effects.size(); j++)
            op->effects[j]->effect_of.push_back(op);
    }
}

void RedBlackHeuristic::dereference_ops() {
    for (int i = 0; i < propositions_rb.size(); i++){
        for (int j = 0; j < propositions_rb[i].size(); j++){
            propositions_rb[i][j].precondition_of = vector<RedBlackOperator *>();
            propositions_rb[i][j].effect_of = vector<RedBlackOperator *>();
        }
    }
    artificial_precondition.precondition_of = vector<RedBlackOperator *>();
}

bool RedBlackHeuristic::is_redblack_plan(const State &state) const {
    
    vector<const Operator*> plan = red_solution_path;
    for (size_t i = 0; i < leaf_factors.size(); ++i){
        for (size_t j = 0; j < black_solution_path[i].size(); ++j){
            plan.push_back(&g_operators[black_solution_path[i][j]]);
        }
    }
    
    vector<int> black_state(g_variable_domain.size(), -1);
    vector<vector<bool> > red_state(g_variable_domain.size());
    
    for (size_t var = 0; var < g_variable_domain.size(); ++var){
        if (assigned_to_factor[var] != -1){ // -1 => red
            black_state[var] = state[var];
        } else {
            red_state[var].resize(g_variable_domain[var], false);
            red_state[var][state[var]] = true;
        }
    }
    
    for (auto const &op : plan){
        for (auto const &pre : op->get_prevail()){
            if (assigned_to_factor[pre.var] != -1){
                if (black_state[pre.var] != pre.prev){
                    cout << "black prevail condition of op not fulfilled: " << op->get_name() << " - " << g_fact_names[pre.var][pre.prev] << endl;;
                    return false;
                }
            } else {
                if (!red_state[pre.var][pre.prev]){
                    cout << "red prevail condition of op not fulfilled: " << op->get_name() << " - " << g_fact_names[pre.var][pre.prev] << endl;;
                    return false;
                }
            }
        }
        for (auto const &pre : op->get_pre_post()){
            if (pre.pre == -1){
                continue;
            }
            if (assigned_to_factor[pre.var] != -1){
                if (black_state[pre.var] != pre.pre){
                    cout << "black precondition of op not fulfilled: " << op->get_name() << " - " << g_fact_names[pre.var][pre.pre] << endl;;
                    return false;
                }
            } else {
                if (!red_state[pre.var][pre.pre]){
                    cout << "red precondition of op not fulfilled: " << op->get_name() << " - " << g_fact_names[pre.var][pre.pre] << endl;
                    return false;
                }
            }
        }
        
        for (auto const &eff : op->get_pre_post()){
            if (assigned_to_factor[eff.var] != -1){
                black_state[eff.var] = eff.post;
            } else {
                red_state[eff.var][eff.post] = true;
            }
        }
    }
    
    for (auto const &goal : g_goal){
        if (assigned_to_factor[goal.first] != -1){
            if (black_state[goal.first] != goal.second){
                cout << "black goal is not fulfilled: " << g_fact_names[goal.first][goal.second] << endl;
                return false;
            }
        } else {
            if (!red_state[goal.first][goal.second]){
                cout << "red goal is not fulfilled: " << g_fact_names[goal.first][goal.second] << endl;
                return false;
            }
        }
    }
    
    return true;
}



/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int RedBlackHeuristic::compute_heuristic(const State &state){
    
#ifdef DEBUG_RB
    cout << "!!!!!!!!!!!!!! computing heuristic for state:" << endl;
    state.dump_pddl();
#endif
    
    if (test_goal(state)){
	cout << "number cache hits " << number_cache_hits << endl;
	return 0;
    }
    
    // initialize blackState
    vector<vector<int> > black_values = vector<vector<int> >(leaf_factors.size());
    for (int var = 0; var < g_variable_domain.size(); var++){
        int affected_factor = assigned_to_factor[var];
	if (affected_factor != -1){
	    black_values[affected_factor].push_back(state[var]);
	}
    }
    for (int i = 0; i < leaf_factors.size(); i++){
        curr_black_state[i] = BlackState(i, black_values[i], this);
    }
    
    if (!leaf_factors.empty() && leaf_factors[0].size() == g_variable_domain.size()){
        if (achieve_black_goals(state)){
            int h = get_black_solution_path_cost();
            
            check_plan_applicability(state);
            
#ifdef DEBUG_RB
            cout << "           reached black goals => found RB plan of length " << h << endl;
#endif
            
#ifdef ONLY_EVAL_INIT_STATE
            exit(0);
#endif

            return h;
        } else {
            return DEAD_END;
        }
    }
    
    if (all_ops_referenced){
        all_ops_referenced = false;
        dereference_ops();
        relaxed_operators = red_relaxed_ops;
        reference_ops();
    }
    
    relaxed_exploration(state);
    
    int h = 0;
    if (red_goals_achieved()){
#ifdef DEBUG_RB
	cout << "		redGoals achieved" << endl;
#endif
	if (black_goals.empty()){
	    h = get_rpg_cost(state);
	    
	    check_plan_applicability(state);
	    
#ifdef DEBUG_RB
	    cout << "		no black goals => found RB plan of length " << h << endl;
#endif
	    
#ifdef ONLY_EVAL_INIT_STATE
	    exit(0);
#endif

	    return h;	// goal reached
	}
	
	if (achieve_black_goals(state)){
	    h = get_black_solution_path_cost() + get_rpg_cost(state);
	    
	    check_plan_applicability(state);
	    
#ifdef DEBUG_RB
	    cout << "		reached black goals => found RB plan of length " << h << endl;
#endif
	    
#ifdef ONLY_EVAL_INIT_STATE
	    exit(0);
#endif

	    return h;
	}
    }
    
    if (black_post_ops.size() == 0){
#ifdef DEBUG_RB
	cout << "		no blackPostOps => dead end" << endl;
#endif

	return DEAD_END;	// nothing new can be achieved from here on
    }
    
    useful_black_post_ops = vector<vector<int> >(leaf_factors.size());
    
    all_ops_referenced = true;
    dereference_ops();
    relaxed_operators = all_relaxed_ops;
    reference_ops();
    
    relaxed_exploration(state);
    
#ifdef DEBUG_RB
    cout << "		second rfp done" << endl;
#endif
    
    if (!red_goals_achieved()){
#ifdef DEBUG_RB
	cout << "		red goals not reachable" << endl;
#endif
        
	return DEAD_END;	// red goals unreachable
    }

#ifdef DEBUG_RB
    cout << "		try to achieve black goals" << endl;
#endif
    
    if (achieve_black_goals(state, true)){	// check if goal reachable
	h = get_black_solution_path_cost() + get_rpg_cost(state);
	
	check_plan_applicability(state);
	
#ifdef DEBUG_RB
	cout << "		black goals reached => found RB plan of length " << h << endl;
#endif
	
#ifdef ONLY_EVAL_INIT_STATE
	exit(0);
#endif
        
	return h;
    }
    
#ifdef DEBUG_RB
    cout << "		goal not reachable => dead end" << endl;
#endif

    return DEAD_END;
}

void RedBlackHeuristic::add_options_to_parser(OptionParser &parser){
	Heuristic::add_options_to_parser(parser);
    
    parser.add_option<int>("max_black_size", "-1 => INT_MAX");
    
    vector<string> preferredTypes = vector<string>();
    preferredTypes.push_back("NONE");
    preferredTypes.push_back("RB");
    preferredTypes.push_back("FF");
    OptionFlags flags;
    flags.mandatory = false;
    parser.add_enum_option("preferred", preferredTypes, "", "NONE", vector<string>(), flags);
}

static Heuristic *_parse(OptionParser &parser) {
    parser.document_synopsis("Red-Black heuristic", "See also Synergy.");
    parser.document_language_support("action costs", "supported");
    parser.document_language_support("conditional_effects", "not supported");
    parser.document_language_support("axioms", "not supported");
    parser.document_property("admissible", "no");
    parser.document_property("consistent", "no");
    parser.document_property("safe", "yes");
    parser.document_property("preferred operators", "yes");

    RedBlackHeuristic::add_options_to_parser(parser);
    Options opts = parser.parse();
    if (parser.dry_run())
        return 0;
    else
        return new RedBlackHeuristic(opts);
}

static Plugin<Heuristic> _plugin("red-black", _parse);
