#ifndef RED_BLACK_HEURISTIC_H
#define RED_BLACK_HEURISTIC_H

#include "ff_heuristic.h"
#include "operator.h"
#include "timer.h"

#include <vector>
#include <deque>
#include <ext/hash_map>

class RedBlackProposition;

enum PreferredOpsType {
    NONE,
    RB,
    FF
};

enum PropositionStatus {
    UNREACHED,
    REACHED
};

struct RedBlackOperator {
    int op_index;
    std::vector<RedBlackProposition *> precondition;
    std::vector<RedBlackProposition *> effects;
    int affects_leaf_factor;

    int cost;
    int unsatisfied_preconditions;
    int h_max_supporter_cost; // h_max_cost of h_max_supporter
    RedBlackProposition *h_max_supporter;
    RedBlackOperator(const std::vector<RedBlackProposition *> &pre,
                    const std::vector<RedBlackProposition *> &eff,
                    int the_op, int cost, int leaf_factor)
        : op_index(the_op), precondition(pre), effects(eff), affects_leaf_factor(leaf_factor),
          cost(cost) {
    }

    inline void update_h_max_supporter();
};

struct RedBlackProposition {
    std::vector<RedBlackOperator *> precondition_of;
    std::vector<RedBlackOperator *> effect_of;

    PropositionStatus status;
    bool marked;
    bool is_goal;
    int h_max_cost;
    int achieved_by;

    RedBlackProposition() {
    }
};

class RedBlackHeuristic : public FFHeuristic {
    
    friend class FFHeuristic;
    
    struct BlackState{
	
    private:
	
        int leaf_factor;
	std::vector<int> facts;
	
    public:
	
	const RedBlackHeuristic *h;     // TODO get rid of this!!
	
	BlackState(int factor, std::vector<int> facts, const RedBlackHeuristic *heuristic) : 
                                            leaf_factor(factor), facts(facts), h(heuristic) {}
	
	bool isApplicable(const Operator &op, const std::vector<std::vector<RedBlackProposition> > &props) const {
	    for (int i = 0; i < op.get_prevail().size(); i++){
		if (h->assigned_to_factor[op.get_prevail()[i].var] != -1){
		    if (facts[h->new_index[op.get_prevail()[i].var]] != op.get_prevail()[i].prev){
			return false;
		    }
		} else {
                    if (props[op.get_prevail()[i].var][op.get_prevail()[i].prev].status != REACHED){
			return false;
		    }
		}
	    }
	    for (int i = 0; i < op.get_pre_post().size(); i++){
		if (op.get_pre_post()[i].pre == -1){
		    continue;
		}
		if (h->assigned_to_factor[op.get_pre_post()[i].var] != -1){
		    if (facts[h->new_index[op.get_pre_post()[i].var]] != op.get_pre_post()[i].pre){
			return false;
		    }
		} else {
                    if (props[op.get_pre_post()[i].var][op.get_pre_post()[i].pre].status != REACHED){
			return false;
		    }
		}
	    }
	    return true;
	}
	
	void apply(const Operator &op){
	    for (int i = 0; i < op.get_pre_post().size(); i++){
		if (h->assigned_to_factor[op.get_pre_post()[i].var] != -1){
		    facts[h->new_index[op.get_pre_post()[i].var]] = op.get_pre_post()[i].post;
		}
	    }
	}
	
	BlackState getSuccessor(const Operator &op) const {
	    std::vector<int> new_facts = facts;
	    for (int i = 0; i < op.get_pre_post().size(); i++){
		if (h->assigned_to_factor[op.get_pre_post()[i].var] != -1){
		    new_facts[h->new_index[op.get_pre_post()[i].var]] = op.get_pre_post()[i].post;
		}
	    }
	    return BlackState(leaf_factor, new_facts, h);
	}
	
	int operator[](int index) const {
	    //std::cout << "size = " << facts.size() << ", index = " << h->new_index[index] << ", leaf = " << leaf_factor << std::endl;
	    return facts[h->new_index[index]];
	}
	
	bool operator<(const BlackState &state) const {
	    int h1 = 1, h2 = 1;
	    for (int i = 0; i < facts.size(); i++){
		h1 = h1 * 37 + facts[i];
		h2 = h2 * 37 + state.facts[i];
	    }
	    return h1 < h2;
	}
	
	void dump() const {
	    std::cout << "BlackState:" << std::endl;
	    for (int i = 0; i < facts.size(); i++){
		std::cout << "   " << g_fact_names[h->leaf_factors[leaf_factor][i]][facts[i]] << std::endl;
	    }
	}
	
    };
    
    struct OpenListEntry {
	
	BlackState state;
	std::deque<int> path;
	
	OpenListEntry(BlackState s, std::deque<int> p) : state(s), path(p){}
	
    };
    
    struct Cache {
	
    private:
	std::vector<std::map<int, std::vector<int> > > cached_paths;
        
        std::vector<int> last_checked_path;
        int last_checked_hash;
        bool last_found;
        int last_checked_factor;
	
    public:
        Cache () {}
	Cache (int number_leaves) : cached_paths(std::vector<std::map<int, std::vector<int> > >(number_leaves)),
                                    last_found(false) {}
	
	void insertPath(int hash, int leaf_factor, std::vector<int> path){
	    cached_paths[leaf_factor].insert(make_pair(hash, path));
	}
	
	bool contains(int hash, int leaf_factor) {
            last_checked_hash = hash;
            last_checked_factor = leaf_factor;
            std::map<int, std::vector<int> >::iterator it = cached_paths[leaf_factor].find(hash);
            if (it != cached_paths[leaf_factor].end()){
                last_checked_path = it->second;
                last_found = true;
                return true;
            } else {
                last_found = false;
                return false;
            }
        }
	
	std::vector<int> get(int hash, int leaf_factor) {
            if (last_found && hash == last_checked_hash && leaf_factor == last_checked_factor){
                return last_checked_path;                
            }
            if (contains(hash, leaf_factor)){
                return last_checked_path;
            } else {
                return std::vector<int>();
            }
	}
	
	
    };
    
    
private:
    std::vector<int> assigned_to_factor;        // -1 => red var
    
    std::vector<std::vector<int> > leaf_factors;
    
    // maps black variables to their index in BlackState, -1 for red vars
    std::vector<int> new_index;
    
    // operators that have black preconditions and effects
    std::vector<std::vector<int> > black_pre_post_ops;
    
    // operators that only have black effects and no black precondition
    std::vector<std::vector<int> > black_post_ops;
    
    // operators only touching the red variables
    std::vector<int> red_ops;
    
    // contains operator indices
    std::vector<std::vector<int> > black_solution_path;
    
    // contains operator indices
    std::vector<const Operator*> red_solution_path;
    
    // those blackPostOps that achieve new red facts in relaxed_exploration
    std::vector<std::vector<int> > useful_black_post_ops;
    
    std::vector<std::vector<std::pair<int, int> > > black_goals;
    
    std::vector<std::pair<int, int> > red_goals;
    
    std::vector<BlackState> curr_black_state;
    
    Cache black_solution_cache;
    
    int number_cache_hits;
    
    std::vector<RedBlackOperator> relaxed_operators;
    std::vector<std::vector<RedBlackProposition> > propositions_rb;
    
    RedBlackProposition artificial_precondition;
    std::vector<RedBlackProposition*> goal_propositions;
    
    AdaptiveQueue<RedBlackProposition *> priority_queue;
    
    std::vector<RedBlackOperator> red_relaxed_ops;
    std::vector<RedBlackOperator> all_relaxed_ops;
    bool all_ops_referenced;
    
    // the maximal size for the black state space
    int max_black_size;
    
    bool solution_found;
    
    PreferredOpsType preferred_ops;
    
    void reference_ops();
    
    void dereference_ops();
    
    bool red_goals_achieved();
    
    //bool all_black_ops_red_pre_achieved();
    
    void insert_red_pre(const Operator &black_op);
    
    int get_hash(const BlackState &state, int leaf_factor);
    
    bool black_goal_reached(const BlackState &curr_state, int leaf_factor);
    
    // checks if operator op is applicable, ignoring black variables
    bool is_applicable_ignore_black(const Operator &op);
    
    // checks if the given operator contains red effects
    bool contains_red_post(const Operator &op);
    
    // returns true if the black goal could be reached, false otherwise
    bool achieve_black_goals(const State &state, bool all_black_ops = false);
    
    // checks if the found RB is a real plan and acts accordingly
    void check_plan_applicability(const State &state);
    
    // returns the cost of the actions found in the relaxedFixedPoint; NOTE: achieveBlackGoals must
    // be called before this function to be able to achieve all red preconditions of the needed black ops
    // also checks if the found red-black plan is actually applicable in the non-relaxed task
    int get_rpg_cost(const State &state);
    
    int get_black_solution_path_cost();
    
    bool is_real_plan(const State &state) const;
    
    bool is_redblack_plan(const State &state) const;
    
    void relaxed_exploration(const State &state);
    
    void build_relaxed_operator(int op_index, int affected_factor);
    
    void enqueue_if_necessary(RedBlackProposition *prop, int cost, const RedBlackOperator *op) {
        assert(cost >= 0);
        if (prop->status == UNREACHED || prop->h_max_cost > cost) {
            prop->status = REACHED;
            prop->h_max_cost = cost;
	    if (op != 0){
		prop->achieved_by = op->op_index;
		if (op->affects_leaf_factor != -1){
		    useful_black_post_ops[op->affects_leaf_factor].push_back(op->op_index);
		}
	    } else {
		prop->achieved_by = -1;
	    }
            priority_queue.push(cost, prop);
        }
    }
    
    
protected:
    virtual void initialize();
    virtual int compute_heuristic(const State &state);
public:
    RedBlackHeuristic(int max_black_size, const Options &options);
    RedBlackHeuristic(const Options &options);
    ~RedBlackHeuristic();
    virtual bool found_solution();
    static void add_options_to_parser(OptionParser &parser);
};

#endif
