#include "scc_counting_anti_heuristic.h"

#include "globals.h"
#include "option_parser.h"
#include "plugin.h"
#include "state.h"
#include "causal_graph.h"
#include "../preprocess/scc.h"

#include <cassert>
#include <stdint.h>
#include <vector>


using namespace std;


// construction and destruction
SCCCountingHeuristic::SCCCountingHeuristic(const Options &opts)
    : Heuristic(opts){
    if (opts.contains("bound")){
	bound = opts.get<int>("bound");
    } else {
	cout << "ERROR: no value for option bound given." << endl;
	exit(EXIT_INPUT_ERROR);
    }
}

SCCCountingHeuristic::~SCCCountingHeuristic() {
}

// initialization
void SCCCountingHeuristic::initialize() {
    cout << "Initializing SCC counting \"heuristic\"..." << endl;
}


int SCCCountingHeuristic::compute_heuristic(const State &/*state*/) {
    
    vector<vector<int> > vars = vector<vector<int> >(g_variable_domain.size());
    for (int i = 0; i < g_variable_domain.size(); i++){
	vars[i] = g_causal_graph->get_successors(i);
    }
    
    SCC scc(vars);
    
    vector<vector<int> > sccs = scc.get_result();
    vector<set<int> > sccVars = vector<set<int> >(sccs.size());
    
    uint64_t SCCSize;
    
    int numberLeafVars = 0;
    
    int numberLeafSccs = 0;
    uint64_t avgLeafSccSize = 0;
    
    int numberLeafSccsWithSucc = 0;
    uint64_t avgLeafSccWithSuccSize = 0;
    int avgNumberSingletonSucc = 0;
    
    bool sccTooBig;
    for (int i = 0; i < sccs.size(); i++){
	if (sccs[i].size() == 1 && g_causal_graph->get_successors(sccs[i][0]).size() == 0){
	    cout << "SCC is CG leaf variable of size " << g_variable_domain[sccs[i][0]] << endl;
	    numberLeafVars++;
	    continue;
	}
	sccTooBig = false;
	sccVars[i] = set<int>(sccs[i].begin(), sccs[i].end());
	cout << "========= 	SCC " << (i+1) << " with " << sccs[i].size() << " variables" << endl;
	cout << "contains variables:" << endl;
	set<int> successorsNotInSCC = set<int>();
	set<int> singletonSucc = set<int>();
	SCCSize = 1;
	for (int j = 0; j < sccs[i].size(); j++){
	    int var = sccs[i][j];
	    SCCSize *= g_variable_domain[var];
	    if (SCCSize > bound){
		sccTooBig = true;
		break;
	    }
	    cout << "	" << g_variable_name[var] << endl;
	    cout << "		";
	    for (int val = 0; val < g_variable_domain[var]; val++){
		cout << " " << g_fact_names[var][val];
	    }
	    cout << endl;
	    for (int succ = 0; succ < g_causal_graph->get_successors(var).size(); succ++){
		int successor = g_causal_graph->get_successors(var)[succ];
		if (sccVars[i].find(successor) == sccVars[i].end()){
		    successorsNotInSCC.insert(successor);
		    if (g_causal_graph->get_successors(successor).size() == 0){
			// successor of current SCC is causal graph leaf
			singletonSucc.insert(successor);
		    }
		}
	    }
	}
	cout << "has size " << SCCSize << endl;
	if (sccTooBig){
	    cout << "SCC size exceeds bound" << endl;
	    continue;
	}
	if (successorsNotInSCC.size() == 0){
	    cout << "========= 	SCC is leaf SCC" << endl;
	    numberLeafSccs++;
	    avgLeafSccSize += SCCSize;
	} else {
	    cout << "========= 	not a leaf SCC" << endl;
	    cout << "	there are " << successorsNotInSCC.size() << " variables, that are CG successors of this SCC" << endl;
	    cout << "	" << singletonSucc.size() << " of them are causal graph leaves" << endl;

	    if (successorsNotInSCC.size() == singletonSucc.size()){
		cout << "===	SCC is an \"extended\" leaf SCC" << endl;
		for (set<int>::const_iterator it = singletonSucc.begin(); it != singletonSucc.end(); ++it){
		    SCCSize *= g_variable_domain[*it];
		    if (SCCSize > bound){
			sccTooBig = true;
			break;
		    }		
		}
		if (sccTooBig){
		    cout << "SCC size exceeds bound" << endl;
		    continue;
		}
		numberLeafSccsWithSucc++;
		avgLeafSccWithSuccSize += SCCSize;
		avgNumberSingletonSucc += singletonSucc.size();
	    }
	    
	    cout << "	singleton successors: 	";
	    for (set<int>::iterator it = singletonSucc.begin(); it != singletonSucc.end(); ++it){
		cout << " " << g_variable_name[*it];
	    }
	    cout << endl;
	    cout << "	all successors: 	";
	    for (set<int>::iterator it = successorsNotInSCC.begin(); it != successorsNotInSCC.end(); ++it){
		cout << " " << g_variable_name[*it];
	    }
	    cout << endl;
	}
    }
    
    cout << "number SCCs " << sccs.size() << endl;
    
    cout << "number leaf variables " << numberLeafVars << endl;
    
    cout << "number leaf SCCs " << numberLeafSccs << endl;
    cout << "average size leaf SCCs " << (numberLeafSccs > 0 ? avgLeafSccSize / numberLeafSccs : 0) << endl;

    cout << "number extended leaf SCCs " << numberLeafSccsWithSucc << endl;
    cout << "average size extended leaf SCCs " << (numberLeafSccsWithSucc > 0 ? avgLeafSccWithSuccSize / numberLeafSccsWithSucc : 0) << endl;
    cout << "average number singleton successors " << (numberLeafSccsWithSucc > 0 ? avgNumberSingletonSucc / numberLeafSccsWithSucc : 0) << endl;
    
    return DEAD_END;
}

void SCCCountingHeuristic::add_options_to_parser(OptionParser &parser){
	Heuristic::add_options_to_parser(parser);
}

static Heuristic *_parse(OptionParser &parser) {
    parser.document_synopsis("SCC counting; no heuristic!", "See also Synergy.");
    parser.document_language_support("action costs", "supported");
    //parser.document_language_support("conditional_effects", "supported");
    parser.document_language_support(
        "axioms",
        "supported (in the sense that the planner won't complain -- "
        "handling of axioms might be very stupid "
        "and even render the heuristic unsafe)");
    parser.document_property("admissible", "no");
    parser.document_property("consistent", "no");
    parser.document_property("safe", "yes for tasks without axioms");
    parser.document_property("preferred operators", "yes");
    
    parser.add_option<int>("bound");

    SCCCountingHeuristic::add_options_to_parser(parser);
    Options opts = parser.parse();
    if (parser.dry_run())
        return 0;
    else
        return new SCCCountingHeuristic(opts);
}

static Plugin<Heuristic> _plugin("scc-count", _parse);
